@@ -2,7 +2,7 @@
 /* vim:set et sts=4: */
 /* ibus - The Input Bus
  * Copyright (C) 2008-2013 Peng Huang <shawn.p.huang@gmail.com>
- * Copyright (C) 2015-2018 Takao Fujiwara <takao.fujiwara1@gmail.com>
+ * Copyright (C) 2015-2022 Takao Fujiwara <takao.fujiwara1@gmail.com>
  * Copyright (C) 2008-2016 Red Hat, Inc.
  *
  * This library is free software; you can redistribute it and/or
@@ -59,6 +59,8 @@ struct _BusEngineProxy {
 
     /* cached properties */
     IBusPropList *prop_list;
+    gboolean has_focus_id;
+    gboolean has_active_surrounding_text;
 };
 
 struct _BusEngineProxyClass {
@@ -105,30 +107,37 @@ static IBusText *text_empty = NULL;
 static IBusPropList *prop_list_empty = NULL;
 
 /* functions prototype */
-static void     bus_engine_proxy_set_property   (BusEngineProxy      *engine,
-                                                 guint                prop_id,
-                                                 const GValue        *value,
-                                                 GParamSpec          *pspec);
-static void     bus_engine_proxy_get_property   (BusEngineProxy      *engine,
-                                                 guint                prop_id,
-                                                 GValue              *value,
-                                                 GParamSpec          *pspec);
+static void     bus_engine_proxy_set_property   (BusEngineProxy    *engine,
+                                                 guint              prop_id,
+                                                 const GValue      *value,
+                                                 GParamSpec        *pspec);
+static void     bus_engine_proxy_get_property   (BusEngineProxy    *engine,
+                                                 guint              prop_id,
+                                                 GValue            *value,
+                                                 GParamSpec        *pspec);
 static void     bus_engine_proxy_real_register_properties
-                                                (BusEngineProxy      *engine,
-                                                 IBusPropList        *prop_list);
+                                                (BusEngineProxy    *engine,
+                                                 IBusPropList      *prop_list);
 static void     bus_engine_proxy_real_update_property
-                                                (BusEngineProxy      *engine,
-                                                 IBusProperty        *prop);
-static void     bus_engine_proxy_real_destroy   (IBusProxy           *proxy);
-static void     bus_engine_proxy_g_signal       (GDBusProxy          *proxy,
-                                                 const gchar         *sender_name,
-                                                 const gchar         *signal_name,
-                                                 GVariant            *parameters);
+                                                (BusEngineProxy    *engine,
+                                                 IBusProperty      *prop);
+static void     bus_engine_proxy_real_destroy   (IBusProxy         *proxy);
+static void     bus_engine_proxy_g_signal       (GDBusProxy        *proxy,
+                                                 const gchar       *sender_name,
+                                                 const gchar       *signal_name,
+                                                 GVariant          *parameters);
 static void     bus_engine_proxy_initable_iface_init
-                                                (GInitableIface      *initable_iface);
+                                                (GInitableIface
+                                                               *initable_iface);
+static void     bus_engine_proxy_get_has_focus_id
+                                                (BusEngineProxy    *engine);
+static void     bus_engine_proxy_get_active_surrounding_text
+                                                (BusEngineProxy    *engine);
 
 G_DEFINE_TYPE_WITH_CODE (BusEngineProxy, bus_engine_proxy, IBUS_TYPE_PROXY,
-                         G_IMPLEMENT_INTERFACE (G_TYPE_INITABLE, bus_engine_proxy_initable_iface_init)
+                         G_IMPLEMENT_INTERFACE (
+                                 G_TYPE_INITABLE,
+                                 bus_engine_proxy_initable_iface_init)
                         );
 
 static GInitableIface *parent_initable_iface = NULL;
@@ -138,8 +147,10 @@ bus_engine_proxy_class_init (BusEngineProxyClass *class)
 {
     GObjectClass *gobject_class = G_OBJECT_CLASS (class);
 
-    gobject_class->set_property = (GObjectSetPropertyFunc)bus_engine_proxy_set_property;
-    gobject_class->get_property = (GObjectGetPropertyFunc)bus_engine_proxy_get_property;
+    gobject_class->set_property =
+            (GObjectSetPropertyFunc)bus_engine_proxy_set_property;
+    gobject_class->get_property =
+            (GObjectGetPropertyFunc)bus_engine_proxy_get_property;
 
     class->register_properties = bus_engine_proxy_real_register_properties;
     class->update_property = bus_engine_proxy_real_update_property;
@@ -147,8 +158,9 @@ bus_engine_proxy_class_init (BusEngineProxyClass *class)
     IBUS_PROXY_CLASS (class)->destroy = bus_engine_proxy_real_destroy;
     G_DBUS_PROXY_CLASS (class)->g_signal = bus_engine_proxy_g_signal;
 
-    parent_initable_iface =
-            (GInitableIface *)g_type_interface_peek (bus_engine_proxy_parent_class, G_TYPE_INITABLE);
+    parent_initable_iface = (GInitableIface *)g_type_interface_peek (
+            bus_engine_proxy_parent_class,
+            G_TYPE_INITABLE);
 
     /* install properties */
     g_object_class_install_property (gobject_class,
@@ -164,7 +176,9 @@ bus_engine_proxy_class_init (BusEngineProxyClass *class)
                         G_PARAM_STATIC_NICK
                         ));
 
-    /* install glib signals that will be sent when corresponding D-Bus signals are sent from an engine process. */
+    /* install glib signals that will be sent when corresponding D-Bus signals
+     * are sent from an engine process.
+     */
     engine_signals[COMMIT_TEXT] =
         g_signal_new (I_("commit-text"),
             G_TYPE_FROM_CLASS (class),
@@ -175,6 +189,9 @@ bus_engine_proxy_class_init (BusEngineProxyClass *class)
             G_TYPE_NONE,
             1,
             IBUS_TYPE_TEXT);
+    g_signal_set_va_marshaller (engine_signals[COMMIT_TEXT],
+                                G_TYPE_FROM_CLASS (class),
+                                bus_marshal_VOID__OBJECTv);
 
     engine_signals[FORWARD_KEY_EVENT] =
         g_signal_new (I_("forward-key-event"),
@@ -188,6 +205,9 @@ bus_engine_proxy_class_init (BusEngineProxyClass *class)
             G_TYPE_UINT,
             G_TYPE_UINT,
             G_TYPE_UINT);
+    g_signal_set_va_marshaller (engine_signals[FORWARD_KEY_EVENT],
+                                G_TYPE_FROM_CLASS (class),
+                                bus_marshal_VOID__UINT_UINT_UINTv);
 
     engine_signals[DELETE_SURROUNDING_TEXT] =
         g_signal_new (I_("delete-surrounding-text"),
@@ -200,6 +220,9 @@ bus_engine_proxy_class_init (BusEngineProxyClass *class)
             2,
             G_TYPE_INT,
             G_TYPE_UINT);
+    g_signal_set_va_marshaller (engine_signals[DELETE_SURROUNDING_TEXT],
+                                G_TYPE_FROM_CLASS (class),
+                                bus_marshal_VOID__INT_UINTv);
 
     engine_signals[REQUIRE_SURROUNDING_TEXT] =
         g_signal_new (I_("require-surrounding-text"),
@@ -210,6 +233,9 @@ bus_engine_proxy_class_init (BusEngineProxyClass *class)
             bus_marshal_VOID__VOID,
             G_TYPE_NONE,
             0);
+    g_signal_set_va_marshaller (engine_signals[REQUIRE_SURROUNDING_TEXT],
+                                G_TYPE_FROM_CLASS (class),
+                                bus_marshal_VOID__VOIDv);
 
     engine_signals[UPDATE_PREEDIT_TEXT] =
         g_signal_new (I_("update-preedit-text"),
@@ -224,6 +250,9 @@ bus_engine_proxy_class_init (BusEngineProxyClass *class)
             G_TYPE_UINT,
             G_TYPE_BOOLEAN,
             G_TYPE_UINT);
+    g_signal_set_va_marshaller (engine_signals[UPDATE_PREEDIT_TEXT],
+                                G_TYPE_FROM_CLASS (class),
+                                bus_marshal_VOID__OBJECT_UINT_BOOLEAN_UINTv);
 
     engine_signals[SHOW_PREEDIT_TEXT] =
         g_signal_new (I_("show-preedit-text"),
@@ -234,6 +263,9 @@ bus_engine_proxy_class_init (BusEngineProxyClass *class)
             bus_marshal_VOID__VOID,
             G_TYPE_NONE,
             0);
+    g_signal_set_va_marshaller (engine_signals[SHOW_PREEDIT_TEXT],
+                                G_TYPE_FROM_CLASS (class),
+                                bus_marshal_VOID__VOIDv);
 
     engine_signals[HIDE_PREEDIT_TEXT] =
         g_signal_new (I_("hide-preedit-text"),
@@ -244,6 +276,9 @@ bus_engine_proxy_class_init (BusEngineProxyClass *class)
             bus_marshal_VOID__VOID,
             G_TYPE_NONE,
             0);
+    g_signal_set_va_marshaller (engine_signals[HIDE_PREEDIT_TEXT],
+                                G_TYPE_FROM_CLASS (class),
+                                bus_marshal_VOID__VOIDv);
 
     engine_signals[UPDATE_AUXILIARY_TEXT] =
         g_signal_new (I_("update-auxiliary-text"),
@@ -256,6 +291,9 @@ bus_engine_proxy_class_init (BusEngineProxyClass *class)
             2,
             IBUS_TYPE_TEXT,
             G_TYPE_BOOLEAN);
+    g_signal_set_va_marshaller (engine_signals[UPDATE_AUXILIARY_TEXT],
+                                G_TYPE_FROM_CLASS (class),
+                                bus_marshal_VOID__OBJECT_BOOLEANv);
 
     engine_signals[SHOW_AUXILIARY_TEXT] =
         g_signal_new (I_("show-auxiliary-text"),
@@ -266,6 +304,9 @@ bus_engine_proxy_class_init (BusEngineProxyClass *class)
             bus_marshal_VOID__VOID,
             G_TYPE_NONE,
             0);
+    g_signal_set_va_marshaller (engine_signals[SHOW_AUXILIARY_TEXT],
+                                G_TYPE_FROM_CLASS (class),
+                                bus_marshal_VOID__VOIDv);
 
     engine_signals[HIDE_AUXILIARY_TEXT] =
         g_signal_new (I_("hide-auxiliary-text"),
@@ -276,6 +317,9 @@ bus_engine_proxy_class_init (BusEngineProxyClass *class)
             bus_marshal_VOID__VOID,
             G_TYPE_NONE,
             0);
+    g_signal_set_va_marshaller (engine_signals[HIDE_AUXILIARY_TEXT],
+                                G_TYPE_FROM_CLASS (class),
+                                bus_marshal_VOID__VOIDv);
 
     engine_signals[UPDATE_LOOKUP_TABLE] =
         g_signal_new (I_("update-lookup-table"),
@@ -288,6 +332,9 @@ bus_engine_proxy_class_init (BusEngineProxyClass *class)
             2,
             IBUS_TYPE_LOOKUP_TABLE,
             G_TYPE_BOOLEAN);
+    g_signal_set_va_marshaller (engine_signals[UPDATE_LOOKUP_TABLE],
+                                G_TYPE_FROM_CLASS (class),
+                                bus_marshal_VOID__OBJECT_BOOLEANv);
 
     engine_signals[SHOW_LOOKUP_TABLE] =
         g_signal_new (I_("show-lookup-table"),
@@ -298,6 +345,9 @@ bus_engine_proxy_class_init (BusEngineProxyClass *class)
             bus_marshal_VOID__VOID,
             G_TYPE_NONE,
             0);
+    g_signal_set_va_marshaller (engine_signals[SHOW_LOOKUP_TABLE],
+                                G_TYPE_FROM_CLASS (class),
+                                bus_marshal_VOID__VOIDv);
 
     engine_signals[HIDE_LOOKUP_TABLE] =
         g_signal_new (I_("hide-lookup-table"),
@@ -308,6 +358,9 @@ bus_engine_proxy_class_init (BusEngineProxyClass *class)
             bus_marshal_VOID__VOID,
             G_TYPE_NONE,
             0);
+    g_signal_set_va_marshaller (engine_signals[HIDE_LOOKUP_TABLE],
+                                G_TYPE_FROM_CLASS (class),
+                                bus_marshal_VOID__VOIDv);
 
     engine_signals[PAGE_UP_LOOKUP_TABLE] =
         g_signal_new (I_("page-up-lookup-table"),
@@ -318,6 +371,9 @@ bus_engine_proxy_class_init (BusEngineProxyClass *class)
             bus_marshal_VOID__VOID,
             G_TYPE_NONE,
             0);
+    g_signal_set_va_marshaller (engine_signals[PAGE_UP_LOOKUP_TABLE],
+                                G_TYPE_FROM_CLASS (class),
+                                bus_marshal_VOID__VOIDv);
 
     engine_signals[PAGE_DOWN_LOOKUP_TABLE] =
         g_signal_new (I_("page-down-lookup-table"),
@@ -328,6 +384,9 @@ bus_engine_proxy_class_init (BusEngineProxyClass *class)
             bus_marshal_VOID__VOID,
             G_TYPE_NONE,
             0);
+    g_signal_set_va_marshaller (engine_signals[PAGE_DOWN_LOOKUP_TABLE],
+                                G_TYPE_FROM_CLASS (class),
+                                bus_marshal_VOID__VOIDv);
 
     engine_signals[CURSOR_UP_LOOKUP_TABLE] =
         g_signal_new (I_("cursor-up-lookup-table"),
@@ -338,6 +397,9 @@ bus_engine_proxy_class_init (BusEngineProxyClass *class)
             bus_marshal_VOID__VOID,
             G_TYPE_NONE,
             0);
+    g_signal_set_va_marshaller (engine_signals[CURSOR_UP_LOOKUP_TABLE],
+                                G_TYPE_FROM_CLASS (class),
+                                bus_marshal_VOID__VOIDv);
 
     engine_signals[CURSOR_DOWN_LOOKUP_TABLE] =
         g_signal_new (I_("cursor-down-lookup-table"),
@@ -348,6 +410,9 @@ bus_engine_proxy_class_init (BusEngineProxyClass *class)
             bus_marshal_VOID__VOID,
             G_TYPE_NONE,
             0);
+    g_signal_set_va_marshaller (engine_signals[CURSOR_DOWN_LOOKUP_TABLE],
+                                G_TYPE_FROM_CLASS (class),
+                                bus_marshal_VOID__VOIDv);
 
     engine_signals[REGISTER_PROPERTIES] =
         g_signal_new (I_("register-properties"),
@@ -359,6 +424,9 @@ bus_engine_proxy_class_init (BusEngineProxyClass *class)
             G_TYPE_NONE,
             1,
             IBUS_TYPE_PROP_LIST);
+    g_signal_set_va_marshaller (engine_signals[REGISTER_PROPERTIES],
+                                G_TYPE_FROM_CLASS (class),
+                                bus_marshal_VOID__OBJECTv);
 
     engine_signals[UPDATE_PROPERTY] =
         g_signal_new (I_("update-property"),
@@ -370,6 +438,9 @@ bus_engine_proxy_class_init (BusEngineProxyClass *class)
             G_TYPE_NONE,
             1,
             IBUS_TYPE_PROPERTY);
+    g_signal_set_va_marshaller (engine_signals[UPDATE_PROPERTY],
+                                G_TYPE_FROM_CLASS (class),
+                                bus_marshal_VOID__OBJECTv);
 
     engine_signals[PANEL_EXTENSION] =
         g_signal_new (I_("panel-extension"),
@@ -377,10 +448,13 @@ bus_engine_proxy_class_init (BusEngineProxyClass *class)
             G_SIGNAL_RUN_LAST,
             0,
             NULL, NULL,
-            bus_marshal_VOID__VARIANT,
+            bus_marshal_VOID__OBJECT,
             G_TYPE_NONE,
             1,
-            G_TYPE_VARIANT);
+            IBUS_TYPE_EXTENSION_EVENT);
+    g_signal_set_va_marshaller (engine_signals[PANEL_EXTENSION],
+                                G_TYPE_FROM_CLASS (class),
+                                bus_marshal_VOID__OBJECTv);
 
     text_empty = ibus_text_new_from_static_string ("");
     g_object_ref_sink (text_empty);
@@ -473,7 +547,8 @@ bus_engine_proxy_real_destroy (IBusProxy *proxy)
         engine->prop_list = NULL;
     }
 
-    IBUS_PROXY_CLASS (bus_engine_proxy_parent_class)->destroy ((IBusProxy *)engine);
+    IBUS_PROXY_CLASS (bus_engine_proxy_parent_class)->destroy (
+            (IBusProxy *)engine);
 }
 
 static void
@@ -486,7 +561,8 @@ _g_object_unref_if_floating (gpointer instance)
 /**
  * bus_engine_proxy_g_signal:
  *
- * Handle all D-Bus signals from the engine process. This function emits corresponding glib signal for the D-Bus signal.
+ * Handle all D-Bus signals from the engine process. This function emits
+ * corresponding glib signal for the D-Bus signal.
  */
 static void
 bus_engine_proxy_g_signal (GDBusProxy  *proxy,
@@ -522,7 +598,9 @@ bus_engine_proxy_g_signal (GDBusProxy  *proxy,
         }
     }
 
-    /* Handle D-Bus signals with parameters. Deserialize them and emit a glib signal. */
+    /* Handle D-Bus signals with parameters. Deserialize them and emit a glib
+     * signal.
+     */
     if (g_strcmp0 (signal_name, "CommitText") == 0) {
         GVariant *arg0 = NULL;
         g_variant_get (parameters, "(v)", &arg0);
@@ -568,7 +646,8 @@ bus_engine_proxy_g_signal (GDBusProxy  *proxy,
         gboolean visible = FALSE;
         guint mode = 0;
 
-        g_variant_get (parameters, "(vubu)", &arg0, &cursor_pos, &visible, &mode);
+        g_variant_get (parameters, "(vubu)",
+                       &arg0, &cursor_pos, &visible, &mode);
         g_return_if_fail (arg0 != NULL);
 
         IBusText *text = IBUS_TEXT (ibus_serializable_deserialize (arg0));
@@ -594,7 +673,11 @@ bus_engine_proxy_g_signal (GDBusProxy  *proxy,
         g_variant_unref (arg0);
         g_return_if_fail (text != NULL);
 
-        g_signal_emit (engine, engine_signals[UPDATE_AUXILIARY_TEXT], 0, text, visible);
+        g_signal_emit (engine,
+                       engine_signals[UPDATE_AUXILIARY_TEXT],
+                       0,
+                       text,
+                       visible);
         _g_object_unref_if_floating (text);
         return;
     }
@@ -606,11 +689,16 @@ bus_engine_proxy_g_signal (GDBusProxy  *proxy,
         g_variant_get (parameters, "(vb)", &arg0, &visible);
         g_return_if_fail (arg0 != NULL);
 
-        IBusLookupTable *table = IBUS_LOOKUP_TABLE (ibus_serializable_deserialize (arg0));
+        IBusLookupTable *table =
+                IBUS_LOOKUP_TABLE (ibus_serializable_deserialize (arg0));
         g_variant_unref (arg0);
         g_return_if_fail (table != NULL);
 
-        g_signal_emit (engine, engine_signals[UPDATE_LOOKUP_TABLE], 0, table, visible);
+        g_signal_emit (engine,
+                       engine_signals[UPDATE_LOOKUP_TABLE],
+                       0,
+                       table,
+                       visible);
         _g_object_unref_if_floating (table);
         return;
     }
@@ -620,11 +708,15 @@ bus_engine_proxy_g_signal (GDBusProxy  *proxy,
         g_variant_get (parameters, "(v)", &arg0);
         g_return_if_fail (arg0 != NULL);
 
-        IBusPropList *prop_list = IBUS_PROP_LIST (ibus_serializable_deserialize (arg0));
+        IBusPropList *prop_list =
+                IBUS_PROP_LIST (ibus_serializable_deserialize (arg0));
         g_variant_unref (arg0);
         g_return_if_fail (prop_list != NULL);
 
-        g_signal_emit (engine, engine_signals[REGISTER_PROPERTIES], 0, prop_list);
+        g_signal_emit (engine,
+                       engine_signals[REGISTER_PROPERTIES],
+                       0,
+                       prop_list);
         _g_object_unref_if_floating (prop_list);
         return;
     }
@@ -634,7 +726,8 @@ bus_engine_proxy_g_signal (GDBusProxy  *proxy,
         g_variant_get (parameters, "(v)", &arg0);
         g_return_if_fail (arg0 != NULL);
 
-        IBusProperty *prop = IBUS_PROPERTY (ibus_serializable_deserialize (arg0));
+        IBusProperty *prop =
+                IBUS_PROPERTY (ibus_serializable_deserialize (arg0));
         g_variant_unref (arg0);
         g_return_if_fail (prop != NULL);
 
@@ -644,7 +737,16 @@ bus_engine_proxy_g_signal (GDBusProxy  *proxy,
     }
 
     if (g_strcmp0 (signal_name, "PanelExtension") == 0) {
-        g_signal_emit (engine, engine_signals[PANEL_EXTENSION], 0, parameters);
+        GVariant *arg0 = NULL;
+        g_variant_get (parameters, "(v)", &arg0);
+        g_return_if_fail (arg0 != NULL);
+
+        IBusExtensionEvent *event = IBUS_EXTENSION_EVENT (
+                ibus_serializable_deserialize (arg0));
+        g_variant_unref (arg0);
+        g_return_if_fail (event != NULL);
+        g_signal_emit (engine, engine_signals[PANEL_EXTENSION], 0, event);
+        _g_object_unref_if_floating (event);
         return;
     }
 
@@ -656,26 +758,62 @@ bus_engine_proxy_new_internal (const gchar     *path,
                                IBusEngineDesc  *desc,
                                GDBusConnection *connection)
 {
+    GDBusProxyFlags flags;
+    BusEngineProxy *engine;
+    BusIBusImpl *ibus = BUS_DEFAULT_IBUS;
+    GHashTable *hash_table = NULL;
+
     g_assert (path);
     g_assert (IBUS_IS_ENGINE_DESC (desc));
     g_assert (G_IS_DBUS_CONNECTION (connection));
 
-    GDBusProxyFlags flags = G_DBUS_PROXY_FLAGS_DO_NOT_AUTO_START;
-    BusEngineProxy *engine =
-        (BusEngineProxy *) g_initable_new (BUS_TYPE_ENGINE_PROXY,
-                                           NULL,
-                                           NULL,
-                                           "desc",              desc,
-                                           "g-connection",      connection,
-                                           "g-interface-name",  IBUS_INTERFACE_ENGINE,
-                                           "g-object-path",     path,
-                                           "g-default-timeout", g_gdbus_timeout,
-                                           "g-flags",           flags,
-                                           NULL);
+    flags = G_DBUS_PROXY_FLAGS_DO_NOT_AUTO_START;
+    engine = (BusEngineProxy *) g_initable_new (
+            BUS_TYPE_ENGINE_PROXY,
+            NULL,
+            NULL,
+            "desc",              desc,
+            "g-connection",      connection,
+            "g-interface-name",  IBUS_INTERFACE_ENGINE,
+            "g-object-path",     path,
+            "g-default-timeout", g_gdbus_timeout,
+            "g-flags",           flags,
+            NULL);
     const gchar *layout = ibus_engine_desc_get_layout (desc);
     if (layout != NULL && layout[0] != '\0') {
         engine->keymap = ibus_keymap_get (layout);
     }
+
+    g_return_val_if_fail (ibus, engine);
+
+    hash_table = bus_ibus_impl_get_engine_focus_id_table (ibus);
+    if (hash_table) {
+        EngineFocusCategory category;
+        category = (EngineFocusCategory)GPOINTER_TO_INT (
+                g_hash_table_lookup (hash_table,
+                                     ibus_engine_desc_get_name (desc)));
+        if (category == ENGINE_FOCUS_CATEGORY_HAS_ID)
+            engine->has_focus_id = TRUE;
+        else if (category == ENGINE_FOCUS_CATEGORY_NO_ID)
+            engine->has_focus_id = FALSE;
+        else
+            bus_engine_proxy_get_has_focus_id (engine);
+    }
+
+    hash_table = bus_ibus_impl_get_engine_active_surrounding_text_table (ibus);
+    if (hash_table) {
+        EngineSurroundingTextCategory category;
+        category = (EngineSurroundingTextCategory)GPOINTER_TO_INT (
+                g_hash_table_lookup (hash_table,
+                                     ibus_engine_desc_get_name (desc)));
+        if (category == ENGINE_SURROUNDING_TEXT_CATEGORY_HAS_ACTIVE)
+            engine->has_active_surrounding_text = TRUE;
+        else if (category == ENGINE_SURROUNDING_TEXT_CATEGORY_NOT_ACTIVE)
+            engine->has_active_surrounding_text = FALSE;
+        else
+            bus_engine_proxy_get_active_surrounding_text (engine);
+    }
+
     return engine;
 }
 
@@ -731,7 +869,8 @@ engine_proxy_new_data_free (EngineProxyNewData *data)
 /**
  * create_engine_ready_cb:
  *
- * A callback function to be called when bus_factory_proxy_create_engine finishes.
+ * A callback function to be called when bus_factory_proxy_create_engine
+ * finishes.
  * Create an BusEngineProxy object and call the GAsyncReadyCallback.
  */
 static void
@@ -766,8 +905,10 @@ create_engine_ready_cb (BusFactoryProxy    *factory,
 /**
  * notify_factory_cb:
  *
- * A callback function to be called when bus_component_start() emits "notify::factory" signal within 5 seconds.
- * Call bus_factory_proxy_create_engine to create the engine proxy asynchronously.
+ * A callback function to be called when bus_component_start() emits
+ * "notify::factory" signal within 5 seconds.
+ * Call bus_factory_proxy_create_engine to create the engine proxy
+ * asynchronously.
  */
 static void
 notify_factory_cb (BusComponent       *component,
@@ -789,22 +930,25 @@ notify_factory_cb (BusComponent       *component,
             data->handler_id = 0;
         }
 
-        /* We *have to* disconnect the cancelled_cb here, since g_dbus_proxy_call
-         * calls create_engine_ready_cb even if the proxy call is cancelled, and
-         * in this case, create_engine_ready_cb itself will return error using
-         * g_task_return_error(). */
+        /* We *have to* disconnect the cancelled_cb here, since
+         * g_dbus_proxy_call calls create_engine_ready_cb even if the proxy
+         * call is cancelled, and in this case, create_engine_ready_cb itself
+         * will return error using g_task_return_error().
+         */
         if (data->cancellable && data->cancelled_handler_id != 0) {
-            g_cancellable_disconnect (data->cancellable, data->cancelled_handler_id);
+            g_cancellable_disconnect (data->cancellable, 
+                                      data->cancelled_handler_id);
             data->cancelled_handler_id = 0;
         }
 
         /* Create engine from factory. */
-        bus_factory_proxy_create_engine (data->factory,
-                                         data->desc,
-                                         data->timeout,
-                                         data->cancellable,
-                                         (GAsyncReadyCallback) create_engine_ready_cb,
-                                         data);
+        bus_factory_proxy_create_engine (
+                data->factory,
+                data->desc,
+                data->timeout,
+                data->cancellable,
+                (GAsyncReadyCallback) create_engine_ready_cb,
+                data);
     }
     /* If factory is NULL, we will continue wait for
      * factory notify signal or timeout */
@@ -813,7 +957,8 @@ notify_factory_cb (BusComponent       *component,
 /**
  * timeout_cb:
  *
- * A callback function to be called when bus_component_start() does not emit "notify::factory" signal within 5 seconds.
+ * A callback function to be called when bus_component_start() does not emit
+ * "notify::factory" signal within 5 seconds.
  * Call the GAsyncReadyCallback and stop the 5 sec timer.
  */
 static gboolean
@@ -918,16 +1063,18 @@ bus_engine_proxy_new (IBusEngineDesc      *desc,
         /* The factory is ready. We'll create the engine proxy directly. */
         g_object_ref (data->factory);
 
-        /* We don't have to connect to cancelled_cb here, since g_dbus_proxy_call
-         * calls create_engine_ready_cb even if the proxy call is cancelled, and
-         * in this case, create_engine_ready_cb itself can return error using
-         * g_task_return_error(). */
-        bus_factory_proxy_create_engine (data->factory,
-                                         data->desc,
-                                         timeout,
-                                         cancellable,
-                                         (GAsyncReadyCallback) create_engine_ready_cb,
-                                         data);
+        /* We don't have to connect to cancelled_cb here, since
+         * g_dbus_proxy_call calls create_engine_ready_cb even if the proxy
+         * call is cancelled, and in this case, create_engine_ready_cb itself
+         * can return error using g_task_return_error().
+         */
+        bus_factory_proxy_create_engine (
+                data->factory,
+                data->desc,
+                timeout,
+                cancellable,
+                (GAsyncReadyCallback) create_engine_ready_cb,
+                data);
     }
 }
 
@@ -969,8 +1116,11 @@ bus_engine_proxy_process_key_event (BusEngineProxy      *engine,
 {
     g_assert (BUS_IS_ENGINE_PROXY (engine));
 
-    if (keycode != 0 && bus_ibus_impl_is_use_sys_layout (BUS_DEFAULT_IBUS) == FALSE) {
-        /* Since use_sys_layout is false, we don't rely on XKB. Try to convert keyval from keycode by using our own mapping. */
+    if (keycode != 0 &&
+        bus_ibus_impl_is_use_sys_layout (BUS_DEFAULT_IBUS) == FALSE) {
+        /* Since use_sys_layout is false, we don't rely on XKB. Try to convert
+         * keyval from keycode by using our own mapping.
+         */
         IBusKeymap *keymap = engine->keymap;
         if (keymap == NULL)
             keymap = BUS_DEFAULT_KEYMAP;
@@ -1133,7 +1283,8 @@ void bus_engine_proxy_set_surrounding_text (BusEngineProxy *engine,
         g_strcmp0 (text->text, engine->surrounding_text->text) != 0 ||
         cursor_pos != engine->surrounding_cursor_pos ||
         anchor_pos != engine->selection_anchor_pos) {
-        GVariant *variant = ibus_serializable_serialize ((IBusSerializable *)text);
+        GVariant *variant =
+                ibus_serializable_serialize ((IBusSerializable *)text);
         if (engine->surrounding_text)
             g_object_unref (engine->surrounding_text);
         engine->surrounding_text = (IBusText *) g_object_ref_sink (text);
@@ -1192,6 +1343,115 @@ bus_engine_proxy_set_content_type (BusEngineProxy *engine,
     g_variant_unref (content_type);
 }
 
+static void
+bus_engine_proxy_get_engine_property (BusEngineProxy     *engine,
+                                      const gchar        *prop_name,
+                                      GAsyncReadyCallback callback,
+                                      GHashTable         *hash_table)
+{
+    g_assert (BUS_IS_ENGINE_PROXY (engine));
+    g_assert (hash_table);
+    g_dbus_proxy_call ((GDBusProxy *) engine,
+                       "org.freedesktop.DBus.Properties.Get",
+                       g_variant_new ("(ss)",
+                                      IBUS_INTERFACE_ENGINE,
+                                      prop_name),
+                       G_DBUS_CALL_FLAGS_NONE,
+                       -1,
+                       NULL,
+                       callback,
+                       g_hash_table_ref (hash_table));
+}
+
+static void
+_get_has_focus_id_cb (GObject        *object,
+                      GAsyncResult   *res,
+                      gpointer        user_data)
+{
+    GHashTable *hash_table = (GHashTable*)user_data;
+    BusEngineProxy *engine;
+    GError *error = NULL;
+    GVariant *result;
+
+    g_return_if_fail (BUS_IS_ENGINE_PROXY (object));
+    engine = BUS_ENGINE_PROXY (object);
+    result = g_dbus_proxy_call_finish (G_DBUS_PROXY (object), res, &error);
+
+    if (result != NULL) {
+        GVariant *variant = NULL;
+        gpointer value;
+        g_variant_get (result, "(v)", &variant);
+        engine->has_focus_id = g_variant_get_boolean (variant);
+        g_variant_unref (variant);
+        g_variant_unref (result);
+        value =  GINT_TO_POINTER (engine->has_focus_id
+                                  ? ENGINE_FOCUS_CATEGORY_HAS_ID
+                                  : ENGINE_FOCUS_CATEGORY_NO_ID);
+        g_hash_table_replace (
+            hash_table,
+            (gpointer)ibus_engine_desc_get_name (engine->desc),
+            value);
+    }
+    g_hash_table_unref (hash_table);
+}
+
+static void
+bus_engine_proxy_get_has_focus_id (BusEngineProxy *engine)
+{
+    BusIBusImpl *ibus = BUS_DEFAULT_IBUS;
+    g_assert (ibus);
+    bus_engine_proxy_get_engine_property (
+            engine,
+            "FocusId",
+            _get_has_focus_id_cb,
+            bus_ibus_impl_get_engine_focus_id_table (ibus));
+}
+
+static void
+_get_active_surrounding_text_cb (GObject        *object,
+                                 GAsyncResult   *res,
+                                 gpointer        user_data)
+{
+    GHashTable *hash_table = (GHashTable*)user_data;
+    BusEngineProxy *engine;
+    GError *error = NULL;
+    GVariant *result;
+
+    g_return_if_fail (BUS_IS_ENGINE_PROXY (object));
+    engine = BUS_ENGINE_PROXY (object);
+    result = g_dbus_proxy_call_finish (G_DBUS_PROXY (object), res, &error);
+
+    if (result != NULL) {
+        GVariant *variant = NULL;
+        gpointer value;
+        g_variant_get (result, "(v)", &variant);
+        engine->has_active_surrounding_text = g_variant_get_boolean (variant);
+        g_variant_unref (variant);
+        g_variant_unref (result);
+        value =  GINT_TO_POINTER (
+                engine->has_active_surrounding_text
+                ? ENGINE_SURROUNDING_TEXT_CATEGORY_HAS_ACTIVE
+                : ENGINE_SURROUNDING_TEXT_CATEGORY_NOT_ACTIVE);
+        g_hash_table_replace (
+            hash_table,
+            (gpointer)ibus_engine_desc_get_name (engine->desc),
+            value);
+    }
+    g_hash_table_unref (hash_table);
+}
+
+static void
+bus_engine_proxy_get_active_surrounding_text (BusEngineProxy *engine)
+{
+    BusIBusImpl *ibus = BUS_DEFAULT_IBUS;
+    g_assert (ibus);
+    bus_engine_proxy_get_engine_property (
+            engine,
+            "ActiveSurroundingText",
+            _get_active_surrounding_text_cb,
+            bus_ibus_impl_get_engine_active_surrounding_text_table (ibus));
+}
+
 /* a macro to generate a function to call a nullary D-Bus method. */
 #define DEFINE_FUNCTION(Name, name)                         \
     void                                                    \
@@ -1214,11 +1474,26 @@ DEFINE_FUNCTION (CursorDown, cursor_down)
 #undef DEFINE_FUNCTION
 
 void
-bus_engine_proxy_focus_in (BusEngineProxy *engine)
+bus_engine_proxy_focus_in (BusEngineProxy *engine,
+                           const gchar    *object_path,
+                           const gchar    *client)
 {
     g_assert (BUS_IS_ENGINE_PROXY (engine));
-    if (!engine->has_focus) {
-        engine->has_focus = TRUE;
+    if (engine->has_focus)
+        return;
+    engine->has_focus = TRUE;
+    if (engine->has_active_surrounding_text)
+        g_signal_emit (engine, engine_signals[REQUIRE_SURROUNDING_TEXT], 0);
+    if (engine->has_focus_id) {
+        g_dbus_proxy_call ((GDBusProxy *)engine,
+                           "FocusInId",
+                           g_variant_new ("(ss)", object_path, client),
+                           G_DBUS_CALL_FLAGS_NONE,
+                           -1,
+                           NULL,
+                           NULL,
+                           NULL);
+    } else {
         g_dbus_proxy_call ((GDBusProxy *)engine,
                            "FocusIn",
                            NULL,
@@ -1231,11 +1506,23 @@ bus_engine_proxy_focus_in (BusEngineProxy *engine)
 }
 
 void
-bus_engine_proxy_focus_out (BusEngineProxy *engine)
+bus_engine_proxy_focus_out (BusEngineProxy *engine,
+                            const gchar    *object_path)
 {
     g_assert (BUS_IS_ENGINE_PROXY (engine));
-    if (engine->has_focus) {
-        engine->has_focus = FALSE;
+    if (!engine->has_focus)
+        return;
+    engine->has_focus = FALSE;
+    if (engine->has_focus_id) {
+        g_dbus_proxy_call ((GDBusProxy *)engine,
+                           "FocusOutId",
+                           g_variant_new ("(s)", object_path),
+                           G_DBUS_CALL_FLAGS_NONE,
+                           -1,
+                           NULL,
+                           NULL,
+                           NULL);
+    } else {
         g_dbus_proxy_call ((GDBusProxy *)engine,
                            "FocusOut",
                            NULL,
@@ -1253,6 +1540,8 @@ bus_engine_proxy_enable (BusEngineProxy *engine)
     g_assert (BUS_IS_ENGINE_PROXY (engine));
     if (!engine->enabled) {
         engine->enabled = TRUE;
+        if (engine->has_active_surrounding_text)
+            g_signal_emit (engine, engine_signals[REQUIRE_SURROUNDING_TEXT], 0);
         g_dbus_proxy_call ((GDBusProxy *)engine,
                            "Enable",
                            NULL,
@@ -1323,6 +1612,44 @@ bus_engine_proxy_is_enabled (BusEngineProxy *engine)
     return engine->enabled;
 }
 
+void
+bus_engine_proxy_panel_extension_received (BusEngineProxy     *engine,
+                                           IBusExtensionEvent *event)
+{
+    GVariant *variant;
+    g_assert (BUS_IS_ENGINE_PROXY (engine));
+    g_assert (IBUS_IS_EXTENSION_EVENT (event));
+
+    variant = ibus_serializable_serialize_object (
+            IBUS_SERIALIZABLE (event));
+    g_return_if_fail (variant != NULL);
+    g_dbus_proxy_call ((GDBusProxy *)engine,
+                       "PanelExtensionReceived",
+                       g_variant_new ("(v)", variant),
+                       G_DBUS_CALL_FLAGS_NONE,
+                       -1,
+                       NULL,
+                       NULL,
+                       NULL);
+}
+
+void
+bus_engine_proxy_panel_extension_register_keys (BusEngineProxy *engine,
+                                                GVariant       *parameters)
+{
+    g_assert (BUS_IS_ENGINE_PROXY (engine));
+    g_assert (parameters);
+
+    g_dbus_proxy_call ((GDBusProxy *)engine,
+                       "PanelExtensionRegisterKeys",
+                       g_variant_new ("(v)", g_variant_ref (parameters)),
+                       G_DBUS_CALL_FLAGS_NONE,
+                       -1,
+                       NULL,
+                       NULL,
+                       NULL);
+}
+
 static gboolean
 initable_init (GInitable     *initable,
                GCancellable  *cancellable,
