@@ -2,8 +2,8 @@
 /* vim:set et sts=4: */
 /* ibus - The Input Bus
  * Copyright (C) 2008-2013 Peng Huang <shawn.p.huang@gmail.com>
- * Copyright (C) 2011-2023 Takao Fujiwara <takao.fujiwara1@gmail.com>
- * Copyright (C) 2008-2021 Red Hat, Inc.
+ * Copyright (C) 2011-2024 Takao Fujiwara <takao.fujiwara1@gmail.com>
+ * Copyright (C) 2008-2024 Red Hat, Inc.
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
@@ -1986,7 +1986,8 @@ _ibus_set_global_shortcut_keys (BusIBusImpl     *ibus,
         }
         ibus->ime_switcher_keys = keys;
         break;
-    default:;
+    default:
+        g_slice_free1 (sizeof (IBusProcessKeyEventData) * (size + 1), keys);
     }
     return TRUE;
 }
@@ -2388,19 +2389,20 @@ bus_ibus_impl_property_changed (BusIBusImpl *service,
                                 GVariant    *value)
 {
     GDBusMessage *message =
-        g_dbus_message_new_signal ("/org/freedesktop/IBus",
+        g_dbus_message_new_signal (IBUS_PATH_IBUS,
                                    "org.freedesktop.DBus.Properties",
                                    "PropertiesChanged");
 
     /* set a non-zero serial to make libdbus happy */
     g_dbus_message_set_serial (message, 1);
-    g_dbus_message_set_sender (message, "org.freedesktop.IBus");
+    g_dbus_message_set_sender (message, IBUS_NAME_OWNER_NAME);
+
 
     GVariantBuilder *builder = g_variant_builder_new (G_VARIANT_TYPE_ARRAY);
     g_variant_builder_add (builder, "{sv}", property_name, value);
     g_dbus_message_set_body (message,
                              g_variant_new ("(sa{sv}as)",
-                                            "org.freedesktop.IBus",
+                                            IBUS_SERVICE_IBUS,
                                             builder,
                                             NULL));
     g_variant_builder_unref (builder);
@@ -2419,12 +2421,12 @@ bus_ibus_impl_emit_signal (BusIBusImpl *ibus,
                            const gchar *signal_name,
                            GVariant    *parameters)
 {
-    GDBusMessage *message = g_dbus_message_new_signal ("/org/freedesktop/IBus",
-                                                       "org.freedesktop.IBus",
+    GDBusMessage *message = g_dbus_message_new_signal (IBUS_PATH_IBUS,
+                                                       IBUS_INTERFACE_IBUS,
                                                        signal_name);
     /* set a non-zero serial to make libdbus happy */
     g_dbus_message_set_serial (message, 1);
-    g_dbus_message_set_sender (message, "org.freedesktop.IBus");
+    g_dbus_message_set_sender (message, IBUS_NAME_OWNER_NAME);
     if (parameters)
         g_dbus_message_set_body (message, parameters);
     bus_dbus_impl_dispatch_message_by_rule (BUS_DEFAULT_DBUS, message, NULL);
@@ -2461,6 +2463,14 @@ bus_ibus_impl_is_embed_preedit_text (BusIBusImpl *ibus)
     return ibus->embed_preedit_text;
 }
 
+gboolean
+bus_ibus_impl_is_use_global_engine (BusIBusImpl *ibus)
+{
+    g_assert (BUS_IS_IBUS_IMPL (ibus));
+
+    return ibus->use_global_engine;
+}
+
 BusInputContext *
 bus_ibus_impl_get_focused_input_context (BusIBusImpl *ibus)
 {
