@@ -68,6 +68,8 @@ ibus_accelerator_valid (guint           keyval,
         IBUS_KEY_Control_R, IBUS_KEY_Meta_L, IBUS_KEY_Meta_R,
         IBUS_KEY_Alt_L, IBUS_KEY_Alt_R, IBUS_KEY_Super_L, IBUS_KEY_Super_R,
         IBUS_KEY_Hyper_L, IBUS_KEY_Hyper_R, IBUS_KEY_ISO_Level3_Shift,
+        IBUS_KEY_ISO_Level3_Lock, IBUS_KEY_ISO_Level5_Shift,
+        IBUS_KEY_ISO_Level5_Lock,
         IBUS_KEY_ISO_Next_Group, IBUS_KEY_ISO_Prev_Group,
         IBUS_KEY_ISO_First_Group, IBUS_KEY_ISO_Last_Group,
         IBUS_KEY_Mode_switch, IBUS_KEY_Num_Lock, IBUS_KEY_Multi_key,
