@@ -1,7 +1,7 @@
 /* -*- mode: C; c-basic-offset: 4; indent-tabs-mode: nil; -*- */
 /* ibus - The Input Bus
  * Copyright (C) 2013-2014 Peng Huang <shawn.p.huang@gmail.com>
- * Copyright (C) 2013-2023 Takao Fujiwara <takao.fujiwara1@gmail.com>
+ * Copyright (C) 2013-2024 Takao Fujiwara <takao.fujiwara1@gmail.com>
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
@@ -192,7 +192,7 @@ parse_compose_sequence (IBusComposeData *compose_data,
 
     for (i = 1; words[i] != NULL; i++) {
         char *start = words[i];
-        char *end = index (words[i], '>');
+        char *end = strchr (start, '>');
         char *match;
         gunichar codepoint;
 
@@ -222,6 +222,12 @@ parse_compose_sequence (IBusComposeData *compose_data,
             compose_data->sequence[n] = codepoint;
         }
 
+        if (codepoint >= 0x10000) {
+            if (!ibus_compose_key_flag (0xffff & codepoint)) {
+                g_warning ("The keysym %s > 0xffff is not supported: %s",
+                           match, line);
+            }
+        }
         if (codepoint == IBUS_KEY_VoidSymbol) {
             g_warning ("Could not get code point of keysym %s: %s",
                        match, line);
@@ -471,7 +477,7 @@ ibus_compose_list_parse_file (const char *compose_file,
             }
             g_free (en_compose);
             if (buf_include.st_ino == buf_parent.st_ino) {
-                g_message ("System en_US Compose is already loaded %s\n",
+                g_message ("System en_US Compose is already loaded %s",
                            include);
                 g_clear_pointer (&include, g_free);
                 continue;
@@ -494,9 +500,9 @@ ibus_compose_list_parse_file (const char *compose_file,
 
 
 static GList *
-ibus_compose_list_check_duplicated (GList  *compose_list,
-                                    int     max_compose_len,
-                                    GSList *compose_tables)
+ibus_compose_list_check_duplicated_with_en (GList  *compose_list,
+                                            int     max_compose_len,
+                                            GSList *compose_tables)
 {
     GList *list;
     static guint *keysyms;
@@ -603,6 +609,7 @@ ibus_compose_list_check_duplicated (GList  *compose_list,
 
 /*
  * Actual typed keysyms have a flag against the definition.
+ * https://gitlab.freedesktop.org/xorg/proto/xorgproto/-/blob/master/include/X11/keysymdef.h?ref_type=heads#L82
  * https://gitlab.freedesktop.org/xorg/lib/libx11/-/blob/master/nls/en_US.UTF-8/Compose.pre#L4559
  */
 guint
@@ -611,11 +618,30 @@ ibus_compose_key_flag (guint key)
     const char *name;
     if (key <= 0xff)
         return 0;
+    /* en-US has MUSICAL SYMBOL compose table */
+    if (key >= 0xd143 && key <= 0xd1e8)
+        return 0x10000;
+    switch (key) {
+    /* <Aogonek> is not used in UTF-8 compose sequences but <ohorn> in EN
+     * compose file and vn keymap is assumed instead.
+     */
+    case 0x1a1:
+    /* <Zabovedot> is not used in UTF-8 compose sequences but <Uhorn> in EN
+     * compose file and vn keymap s assumed instead.
+     */
+    case 0x1af:
+    /* <caron> is not used in UTF-8 compose sequences but <EZH> in EN compose
+     * file and fr(nodeadkeys) keymap is assumed instead.
+     */
+    case 0x1b7:
+        return 0x1000000;
+    default:;
+    }
     name = ibus_keyval_name (key);
     /* If name is null, the key sequence is expressed as "<Uxxxx>" format in
      * the Compose file and the typed keysym has the flag.
      */
-    if (!name)
+    if (!name || g_str_has_prefix (name, "0x"))
         return 0x1000000;
     /* "<Pointer_EnableKeys>" is not described in the Compose file but <UFEF9>
      * in the file.
@@ -648,8 +674,8 @@ ibus_compose_data_compare (gpointer a,
         gunichar code_a = compose_data_a->sequence[i];
         gunichar code_b = compose_data_b->sequence[i];
 
-        code_a += ibus_compose_key_flag (code_a);
-        code_b += ibus_compose_key_flag (code_b);
+        code_a &= 0xffff;
+        code_b &= 0xffff;
         if (code_a != code_b)
             return code_a - code_b;
         if (code_a == 0 && code_b == 0)
@@ -659,6 +685,56 @@ ibus_compose_data_compare (gpointer a,
 }
 
 
+static GList *
+ibus_compose_list_check_duplicated_with_own (GList  *compose_list,
+                                             int     max_compose_len)
+{
+    GList *list;
+    IBusComposeData *compose_data_a, *compose_data_b;
+    int i;
+    GList *removed_list = NULL;
+
+    for (list = compose_list; list != NULL; list = list->next) {
+        gboolean is_different_value = FALSE;
+        if (!list->next)
+            break;
+        if (!ibus_compose_data_compare (list->data,
+                                        list->next->data,
+                                        GINT_TO_POINTER (max_compose_len))) {
+            compose_data_a = list->data;
+            compose_data_b = list->next->data;
+            for (i = 0; compose_data_a->values[i]; i++) {
+                if (compose_data_a->values[i] != compose_data_b->values[i]) {
+                    is_different_value = TRUE;
+                    break;
+                }
+            }
+            if (is_different_value)
+                g_warning ("Deleting different outputs for same sequence.");
+            else
+                g_debug ("Deleting same compose output for same sequence.");
+            g_print ("{");
+            for (i = 0; compose_data_a->values[i]; i++)
+                g_print ("U+%X, ", compose_data_a->values[i]);
+            g_print ("}");
+            g_print (" {");
+            for (i = 0; compose_data_b->values[i]; i++)
+                g_print ("U+%X, ", compose_data_b->values[i]);
+            g_print ("}\n");
+            removed_list = g_list_append (removed_list, compose_data_a);
+        }
+    }
+    for (list = removed_list; list != NULL; list = list->next) {
+        compose_data_a = list->data;
+        compose_list = g_list_remove (compose_list, compose_data_a);
+        ibus_compose_data_free (compose_data_a);
+    }
+
+    g_list_free (removed_list);
+    return compose_list;
+}
+
+
 static void
 ibus_compose_list_print (GList *compose_list,
                          int    max_compose_len,
@@ -773,7 +849,7 @@ compose_data_to_variant (gconstpointer compose_data,
     g_assert (compose_data);
     if (error)
         *error = NULL;
-    if ((index_stride * n_seqs) > G_MAXUINT64) {
+    if (n_seqs == 0 || index_stride > (G_MAXSIZE / n_seqs)) {
         if (error) {
             g_set_error (error, IBUS_ERROR, IBUS_ERROR_FAILED,
                          "Length %u x %lu is too long",
@@ -1328,6 +1404,7 @@ ibus_compose_table_new_with_list (GList   *compose_list,
                         (G_MAXSIZE / sizeof (guint16)))) {
             g_warning ("Too long allocation %lu x %u",
                        s_size_total - s_size_16bit, n_index_stride);
+            g_free (ibus_compose_seqs);
             return NULL;
         }
         rawdata = (gpointer)g_new (
@@ -1340,6 +1417,8 @@ ibus_compose_table_new_with_list (GList   *compose_list,
                        s_size_total - s_size_16bit,
                        n_index_stride,
                        v_size_32bit);
+            g_free (ibus_compose_seqs);
+            g_free (rawdata);
             return NULL;
         }
         if (G_LIKELY (rawdata)) {
@@ -1356,6 +1435,7 @@ ibus_compose_table_new_with_list (GList   *compose_list,
         }
         if (!ibus_compose_seqs_32bit_first || !ibus_compose_seqs_32bit_second) {
             g_warning ("Failed g_new");
+            g_free (ibus_compose_seqs);
             g_free (rawdata);
             return NULL;
         }
@@ -1447,14 +1527,18 @@ ibus_compose_table_new_with_file (const gchar *compose_file,
     if (compose_list == NULL)
         return NULL;
     n_index_stride = max_compose_len + 2;
-    compose_list = ibus_compose_list_check_duplicated (compose_list,
-                                                       max_compose_len,
-                                                       compose_tables);
+    compose_list = ibus_compose_list_check_duplicated_with_en (compose_list,
+                                                               max_compose_len,
+                                                               compose_tables);
     compose_list = g_list_sort_with_data (
             compose_list,
             (GCompareDataFunc) ibus_compose_data_compare,
             GINT_TO_POINTER (max_compose_len));
 
+    compose_list = ibus_compose_list_check_duplicated_with_own (
+            compose_list,
+            max_compose_len);
+
     if (compose_list == NULL) {
         g_warning ("compose file %s does not include any keys besides keys "
                    "in en-us compose file", compose_file);
@@ -1581,9 +1665,9 @@ compare_seq (const void *key, const void *value)
         guint saved_key = (guint)seq[i];
         guint flag = ibus_compose_key_flag (saved_key);
         if (typed_key < (saved_key + flag))
-            return -1;
+            return (0xffff & typed_key) - saved_key;
         else if (typed_key > (saved_key + flag))
-            return 1;
+            return (0xffff & typed_key) - saved_key;
 
         i++;
     }
