@@ -2,7 +2,7 @@
 /* vim:set et sts=4: */
 /* ibus - The Input Bus
  * Copyright (C) 2014 Peng Huang <shawn.p.huang@gmail.com>
- * Copyright (C) 2015-2023 Takao Fujiwara <takao.fujiwara1@gmail.com>
+ * Copyright (C) 2015-2024 Takao Fujiwara <takao.fujiwara1@gmail.com>
  * Copyright (C) 2014-2017 Red Hat, Inc.
  *
  * This library is free software; you can redistribute it and/or
@@ -83,11 +83,13 @@ struct _IBusEngineSimplePrivate {
     IBusEngineDict     *emoji_dict;
     IBusLookupTable    *lookup_table;
     gboolean            lookup_table_visible;
+    IBusText           *updated_preedit;
 };
 
 guint COMPOSE_BUFFER_SIZE = 20;
 G_LOCK_DEFINE_STATIC (global_tables);
 static GSList *global_tables;
+static IBusText *updated_preedit_empty;
 
 /* functions prototype */
 static void     ibus_engine_simple_destroy      (IBusEngineSimple   *simple);
@@ -135,24 +137,29 @@ ibus_engine_simple_class_init (IBusEngineSimpleClass *class)
     engine_class->page_up   = ibus_engine_simple_page_up;
     engine_class->candidate_clicked
                             = ibus_engine_simple_candidate_clicked;
+    updated_preedit_empty = ibus_text_new_from_string ("");
+    g_object_ref_sink (updated_preedit_empty);
 }
 
 static void
 ibus_engine_simple_init (IBusEngineSimple *simple)
 {
+    IBusEngineSimplePrivate *priv;
     GBytes *data;
     GError *error = NULL;
     const char *contents;
     gsize length = 0;
     IBusComposeTableEx *en_compose_table;
 
-    simple->priv = IBUS_ENGINE_SIMPLE_GET_PRIVATE (simple);
-    simple->priv->compose_buffer = g_new0 (guint, COMPOSE_BUFFER_SIZE + 1);
-    simple->priv->hex_mode_enabled =
+    priv = simple->priv = IBUS_ENGINE_SIMPLE_GET_PRIVATE (simple);
+    priv->compose_buffer = g_new0 (guint, COMPOSE_BUFFER_SIZE + 1);
+    priv->hex_mode_enabled =
         g_getenv("IBUS_ENABLE_CTRL_SHIFT_U") != NULL ||
         g_getenv("IBUS_ENABLE_CONTROL_SHIFT_U") != NULL;
-    simple->priv->tentative_match = g_string_new ("");
-    simple->priv->tentative_match_len = 0;
+    priv->tentative_match = g_string_new ("");
+    priv->tentative_match_len = 0;
+    priv->updated_preedit =
+            (IBusText *)g_object_ref_sink (updated_preedit_empty);
     data = g_resources_lookup_data ("/org/freedesktop/ibus/compose/sequences",
                                     G_RESOURCE_LOOKUP_FLAGS_NONE,
                                     &error);
@@ -190,6 +197,7 @@ ibus_engine_simple_destroy (IBusEngineSimple *simple)
     g_string_free (priv->tentative_match, TRUE);
     priv->tentative_match = NULL;
     priv->tentative_match_len = 0;
+    g_clear_object (&priv->updated_preedit);
 
     IBUS_OBJECT_CLASS(ibus_engine_simple_parent_class)->destroy (
         IBUS_OBJECT (simple));
@@ -228,6 +236,9 @@ ibus_engine_simple_reset (IBusEngine *engine)
         priv->tentative_match_len = 0;
     }
     ibus_engine_hide_preedit_text ((IBusEngine *)simple);
+    g_object_unref (priv->updated_preedit);
+    priv->updated_preedit =
+            (IBusText *)g_object_ref_sink (updated_preedit_empty);
 }
 
 static void
@@ -318,7 +329,10 @@ ibus_engine_simple_update_preedit_text (IBusEngineSimple *simple)
         int len = strlen (priv->tentative_emoji);
         ibus_text_append_attribute (text,
                 IBUS_ATTR_TYPE_UNDERLINE, IBUS_ATTR_UNDERLINE_SINGLE, 0, len);
+        g_object_ref_sink (text);
         ibus_engine_update_preedit_text ((IBusEngine *)simple, text, len, TRUE);
+        g_object_unref (priv->updated_preedit);
+        priv->updated_preedit = text;
         g_string_free (s, TRUE);
         return;
     } else if (priv->in_compose_sequence) {
@@ -357,8 +371,15 @@ ibus_engine_simple_update_preedit_text (IBusEngineSimple *simple)
                     }
                 } else {
                     ch = ibus_keyval_to_unicode (keysym);
-                    if (ch)
+                    if (ch) {
                         g_string_append_unichar(s, ch);
+                    /* Can send Unicode char as keysym with <Uxxxx> format
+                     * in comopse sequences and should not warn this case.
+                     */
+                    } else if (g_unichar_validate (keysym)) {
+                        ch = keysym;
+                        g_string_append_unichar(s, ch);
+                    }
                 }
                 if (!ch) {
                     g_warning (
@@ -370,7 +391,19 @@ ibus_engine_simple_update_preedit_text (IBusEngineSimple *simple)
     }
 
     if (s->len == 0) {
-        ibus_engine_hide_preedit_text ((IBusEngine *)simple);
+        /* #2536 IBusEngine can inherit IBusEngineSimple for comopse keys.
+         * If the previous preedit is zero, the current preedit does not
+         * need to be hidden here at least because ibus-daemon could have
+         * another preedit for the child IBusEnigne likes m17n and caclling
+         * ibus_engine_hide_preedit_text() here could cause a reset of
+         * the cursor position in ibus-daemon.
+         */
+        if (strlen (priv->updated_preedit->text)) {
+            ibus_engine_hide_preedit_text ((IBusEngine *)simple);
+            g_object_unref (priv->updated_preedit);
+            priv->updated_preedit =
+                    (IBusText *)g_object_ref_sink (updated_preedit_empty);
+        }
     } else if (s->len >= G_MAXINT) {
         g_warning ("%s is too long compose length: %lu", s->str, s->len);
     } else {
@@ -378,7 +411,10 @@ ibus_engine_simple_update_preedit_text (IBusEngineSimple *simple)
         IBusText *text = ibus_text_new_from_string (s->str);
         ibus_text_append_attribute (text,
                 IBUS_ATTR_TYPE_UNDERLINE, IBUS_ATTR_UNDERLINE_SINGLE, 0, len);
+        g_object_ref_sink (text);
         ibus_engine_update_preedit_text ((IBusEngine *)simple, text, len, TRUE);
+        g_object_unref (priv->updated_preedit);
+        priv->updated_preedit = text;
     }
     g_string_free (s, TRUE);
 }
@@ -1017,7 +1053,9 @@ ibus_engine_simple_process_key_event (IBusEngine *engine,
          !is_hex_start && !is_hex_end && !is_escape && !is_backspace) ||
         (priv->in_emoji_sequence && !printable_keyval &&
          !is_emoji_start && !is_hex_end && !is_escape && !is_backspace)) {
-        if (modifiers & (IBUS_MOD1_MASK | IBUS_CONTROL_MASK) ||
+        guint no_text_input_mask = IBUS_MOD1_MASK | IBUS_MOD4_MASK \
+                                   | IBUS_CONTROL_MASK | IBUS_SUPER_MASK;
+        if (modifiers & no_text_input_mask ||
             ((priv->in_hex_sequence || priv->in_emoji_sequence) &&
              priv->modifiers_dropped &&
              (keyval == IBUS_KEY_Return ||
