@@ -1,29 +1,33 @@
 /* -*- mode: C; c-basic-offset: 4; indent-tabs-mode: nil; -*- */
 /* vim:set et sts=4: */
 /* ibus - The Input Bus
- * Copyright (C) 2008-2010 Peng Huang <shawn.p.huang@gmail.com>
- * Copyright (C) 2008-2010 Red Hat, Inc.
+ * Copyright (C) 2008-2013 Peng Huang <shawn.p.huang@gmail.com>
+ * Copyright (C) 2015-2022 Takao Fujiwara <takao.fujiwara1@gmail.com>
+ * Copyright (C) 2008-2022 Red Hat, Inc.
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
  * License as published by the Free Software Foundation; either
- * version 2 of the License, or (at your option) any later version.
+ * version 2.1 of the License, or (at your option) any later version.
  *
  * This library is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the GNU
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  * Lesser General Public License for more details.
  *
  * You should have received a copy of the GNU Lesser General Public
- * License along with this library; if not, write to the
- * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
- * Boston, MA 02111-1307, USA.
+ * License along with this library; if not, write to the Free Software
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
+ * USA
  */
 
 #if !defined (__IBUS_H_INSIDE__) && !defined (IBUS_COMPILATION)
 #error "Only <ibus.h> can be included directly"
 #endif
 
+#ifndef __IBUS_SHARE_H_
+#define __IBUS_SHARE_H_
+
 /**
  * SECTION: ibusshare
  * @short_description: Shared utility functions and definition.
@@ -33,11 +37,14 @@
  * which are shared among ibus component and services.
  */
 
-#ifndef __IBUS_SHARE_H_
-#define __IBUS_SHARE_H_
-
 #include <glib.h>
 
+#ifdef IBUS_DISABLE_DEPRECATION_WARNINGS
+#define IBUS_DEPRECATED
+#else
+#define IBUS_DEPRECATED G_DEPRECATED
+#endif
+
 /**
  * IBUS_SERVICE_IBUS:
  *
@@ -45,6 +52,13 @@
  */
 #define IBUS_SERVICE_IBUS       "org.freedesktop.IBus"
 
+/**
+ * IBUS_SERVICE_PORTAL:
+ *
+ * Address of IBus portalservice.
+ */
+#define IBUS_SERVICE_PORTAL     "org.freedesktop.portal.IBus"
+
 /**
  * IBUS_SERVICE_PANEL:
  *
@@ -52,6 +66,22 @@
  */
 #define IBUS_SERVICE_PANEL      "org.freedesktop.IBus.Panel"
 
+/**
+ * IBUS_SERVICE_PANEL_EXTENSION:
+ *
+ * Address of IBus panel extension service.
+ */
+#define IBUS_SERVICE_PANEL_EXTENSION "org.freedesktop.IBus.Panel.Extension"
+
+/**
+ * IBUS_SERVICE_PANEL_EXTENSION_EMOJI:
+ *
+ * Address of IBus panel extension service for emoji.
+ * This service provides emoji, Unicode code point, Unicode name features.
+ */
+#define IBUS_SERVICE_PANEL_EXTENSION_EMOJI \
+        "org.freedesktop.IBus.Panel.Extension.Emoji"
+
 /**
  * IBUS_SERVICE_CONFIG:
  *
@@ -87,6 +117,15 @@
  */
 #define IBUS_PATH_PANEL         "/org/freedesktop/IBus/Panel"
 
+/**
+ * IBUS_PATH_PANEL_EXTENSION_EMOJI:
+ *
+ * D-Bus path for IBus extension panel for emoji.
+ * This service provides emoji, Unicode code point, Unicode name features.
+ */
+#define IBUS_PATH_PANEL_EXTENSION_EMOJI \
+        "/org/freedesktop/IBus/Panel/Extension/Emoji"
+
 /**
  * IBUS_PATH_CONFIG:
  *
@@ -115,6 +154,13 @@
  */
 #define IBUS_INTERFACE_IBUS     "org.freedesktop.IBus"
 
+/**
+ * IBUS_INTERFACE_PORTAL:
+ *
+ * D-Bus interface for IBus portal.
+ */
+#define IBUS_INTERFACE_PORTAL   "org.freedesktop.IBus.Portal"
+
 /**
  * IBUS_INTERFACE_INPUT_CONTEXT:
  *
@@ -162,9 +208,10 @@ G_BEGIN_DECLS
 
 /**
  * ibus_get_local_machine_id:
- * @returns: A newly allocated string that shows the UUID of the machine.
  *
  * Obtains the machine UUID of the machine this process is running on.
+ *
+ * Returns: A newly allocated string that shows the UUID of the machine.
  */
 const gchar     *ibus_get_local_machine_id
                                         (void);
@@ -179,7 +226,6 @@ void             ibus_set_display       (const gchar    *display);
 
 /**
  * ibus_get_address:
- * @returns: D-Bus address of IBus. %NULL for not found.
  *
  * Return the D-Bus address of IBus.
  * It will find the address from following source:
@@ -188,7 +234,9 @@ void             ibus_set_display       (const gchar    *display);
  *    <listitem><para>Socket file under ~/.config/ibus/bus/</para></listitem>
  * </orderedlist>
  *
- * @see_also: ibus_write_address().
+ * Returns: D-Bus address of IBus. %NULL for not found.
+ *
+ * See also: ibus_write_address().
  */
 const gchar     *ibus_get_address       (void);
 
@@ -198,13 +246,12 @@ const gchar     *ibus_get_address       (void);
  *
  * Write D-Bus address to socket file.
  *
- * @see_also: ibus_get_address().
+ * See also: ibus_get_address().
  */
 void             ibus_write_address     (const gchar    *address);
 
 /**
  * ibus_get_user_name:
- * @returns: A newly allocated string that stores current user name.
  *
  * Get the current user name.
  * It is determined by:
@@ -217,15 +264,18 @@ void             ibus_write_address     (const gchar    *address);
  *    <listitem><para>Environment variable USER</para></listitem>
  *    <listitem><para>Environment variable LNAME</para></listitem>
  * </orderedlist>
+ *
+ * Returns: A newly allocated string that stores current user name.
  */
 const gchar     *ibus_get_user_name     (void);
 
 /**
  * ibus_get_daemon_uid:
- * @returns: UID of ibus-daemon; or 0 if UID is not available.
  *
  * Get UID of ibus-daemon.
  *
+ * Returns: UID of ibus-daemon; or 0 if UID is not available.
+ *
  * Deprecated: This function has been deprecated and should
  * not be used in newly written code.
  */
@@ -233,41 +283,26 @@ glong            ibus_get_daemon_uid    (void) G_GNUC_DEPRECATED;
 
 /**
  * ibus_get_socket_path:
- * @returns: A newly allocated string that stores the path of socket file.
  *
  * Get the path of socket file.
+ *
+ * Returns: A newly allocated string that stores the path of socket file.
  */
 const gchar     *ibus_get_socket_path   (void);
 
 /**
  * ibus_get_timeout:
- * @returns: A GDBus timeout in milliseconds. -1 when default timeout for GDBus should be used.
  *
- * Get the GDBus timeout in milliseconds. The timeout is for clients (e.g. im-ibus.so), not for ibus-daemon.
- * Note that the timeout for ibus-daemon could be set by --timeout command line option of the daemon.
+ * Get the GDBus timeout in milliseconds. The timeout is for clients (e.g.
+ * im-ibus.so), not for ibus-daemon.
+ * Note that the timeout for ibus-daemon could be set by --timeout command
+ * line option of the daemon.
+ *
+ * Returns: A GDBus timeout in milliseconds. -1 when default timeout for
+ *     GDBus should be used.
  */
 gint             ibus_get_timeout       (void);
 
-/**
- * ibus_keyval_name:
- * @keyval: Key symbol.
- * @returns: Corresponding key name. %NULL if no such key symbol.
- *
- * Return the name of a key symbol.
- *
- * Note that the returned string is used internally, so don't free it.
- */
-const gchar     *ibus_keyval_name       (guint           keyval);
-
-/**
- * ibus_keyval_from_name:
- * @keyval_name: Key name in #gdk_keys_by_name.
- * @returns: Corresponding key symbol.
- *
- * Return the key symbol that associate with the key name.
- */
-guint            ibus_keyval_from_name  (const gchar    *keyval_name);
-
 /**
  * ibus_free_strv:
  * @strv: List of strings.
@@ -283,25 +318,27 @@ void             ibus_free_strv         (gchar          **strv)
  * ibus_key_event_to_string:
  * @keyval: Key symbol.
  * @modifiers: Modifiers such as Ctrl or Shift.
- * @returns: The name of a key symbol and modifier.
  *
  * Return the name of a key symbol and modifiers.
  *
  * For example, if press ctrl, shift, and enter, then this function returns:
  * Shift+Control+enter.
+ *
+ * Returns: The name of a key symbol and modifier.
  */
-const gchar     *ibus_key_event_to_string
+gchar           *ibus_key_event_to_string
                                         (guint           keyval,
                                          guint           modifiers);
 
 /**
  * ibus_key_event_from_string:
  * @string: Key event string.
- * @keyval: Variable that hold key symbol result.
- * @modifiers: Variable that hold modifiers result.
- * @returns: TRUE for succeed; FALSE if failed.
+ * @keyval: (out): Variable that hold key symbol result.
+ * @modifiers: (out): Variable that hold modifiers result.
  *
  * Parse key event string and return key symbol and modifiers.
+ *
+ * Returns: %TRUE for succeed; %FALSE if failed.
  */
 gboolean         ibus_key_event_from_string
                                         (const gchar    *string,
@@ -311,9 +348,7 @@ gboolean         ibus_key_event_from_string
 /**
  * ibus_init:
  *
- * Init the ibus types.
- *
- * It is actually a wrapper of g_type_init().
+ * Initialize the ibus types.
  */
 void             ibus_init              (void);
 
@@ -322,7 +357,7 @@ void             ibus_init              (void);
  *
  * Runs an IBus main loop until ibus_quit() is called in the loop.
  *
- * @see_also: ibus_quit().
+ * See also: ibus_quit().
  */
 void             ibus_main              (void);
 
@@ -332,7 +367,7 @@ void             ibus_main              (void);
  * Stops an IBus from running.
  *
  * Any calls to ibus_quit() for the loop will return.
- * @see_also: ibus_main().
+ * See also: ibus_main().
  */
 void             ibus_quit              (void);
 
@@ -345,7 +380,7 @@ void             ibus_quit              (void);
  *
  * (ibus-daemon:7088): IBUS-DEBUG: 18:06:45.822819: ibus-daemon started
  *
- * If @verbose is TRUE, all levels of messages will be logged. Otherwise,
+ * If @verbose is %TRUE, all levels of messages will be logged. Otherwise,
  * DEBUG and WARNING messages will be ignored.  The function is used in
  * ibus-daemon, but can be useful for IBus client programs as well for
  * debugging. It's totally fine for not calling this function. If you
@@ -354,5 +389,12 @@ void             ibus_quit              (void);
  */
 void             ibus_set_log_handler   (gboolean verbose);
 
+/**
+ * ibus_unset_log_handler:
+ *
+ * Remove the log handler which is set by ibus_set_log_handler.
+ */
+void             ibus_unset_log_handler (void);
+
 G_END_DECLS
 #endif
